@@ -80,6 +80,10 @@ const gchar *lightdm_greeter_get_default_session_hint (LightDMGreeter *greeter);
 
 gboolean lightdm_greeter_get_hide_users_hint (LightDMGreeter *greeter);
 
+gboolean lightdm_greeter_get_show_manual_login_hint (LightDMGreeter *greeter);
+
+gboolean lightdm_greeter_get_lock_hint (LightDMGreeter *greeter);
+
 gboolean lightdm_greeter_get_has_guest_account_hint (LightDMGreeter *greeter);
 
 const gchar *lightdm_greeter_get_select_user_hint (LightDMGreeter *greeter);
