@@ -27,8 +27,8 @@
 #include "seat-xdmcp-session.h"
 #include "seat-xvnc.h"
 #include "xserver.h"
-#include "pam-session.h"
 #include "process.h"
+#include "session-child.h"
 
 static gchar *config_path = NULL;
 static GMainLoop *loop = NULL;
@@ -239,37 +239,6 @@ set_seat_properties (Seat *seat, const gchar *config_section)
     }
 }
 
-static Session *
-get_session_for_cookie (const gchar *cookie, Seat **seat)
-{
-    GList *link;
- 
-    for (link = display_manager_get_seats (display_manager); link; link = link->next)
-    {
-        Seat *s = link->data;
-        GList *l;
-         
-        for (l = seat_get_displays (s); l; l = l->next)
-        {
-            Display *display = l->data;
-            Session *session;
-
-            session = display_get_session (display);
-            if (!session)
-                continue;
-
-            if (g_strcmp0 (session_get_console_kit_cookie (session), cookie) == 0)
-            {
-                if (seat)
-                    *seat = s;
-                return session;
-            }
-        }
-    }
-
-    return NULL;
-}
-
 static void
 handle_display_manager_call (GDBusConnection       *connection,
                              const gchar           *sender,
@@ -362,47 +331,6 @@ handle_display_manager_call (GDBusConnection       *connection,
             g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED, "Failed to start seat");
         g_object_unref (seat);
     }
-    /* NOTE: This method is deprecated, use the XSG_SEAT_PATH environment variable instead */
-    else if (g_strcmp0 (method_name, "GetSeatForCookie") == 0)
-    {       
-        gchar *cookie;
-        Seat *seat = NULL;
-        BusEntry *entry = NULL;
-
-        if (!g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(s)")))
-            return;
-
-        g_variant_get (parameters, "(&s)", &cookie);
-
-        get_session_for_cookie (cookie, &seat);
-        if (seat)
-            entry = g_hash_table_lookup (seat_bus_entries, seat);
-        if (entry)
-            g_dbus_method_invocation_return_value (invocation, g_variant_new ("(o)", entry->path));
-        else // FIXME: Need to make proper error
-            g_dbus_method_invocation_return_error_literal (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED, "Unable to find seat for cookie");
-    }
-
-    /* NOTE: This method is deprecated, use the XSG_SESSION_PATH environment variable instead */
-    else if (g_strcmp0 (method_name, "GetSessionForCookie") == 0)
-    {      
-        gchar *cookie;
-        Session *session;
-        BusEntry *entry = NULL;
-
-        if (!g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(s)")))
-            return;
-
-        g_variant_get (parameters, "(&s)", &cookie);
-
-        session = get_session_for_cookie (cookie, NULL);
-        if (session)
-            entry = g_hash_table_lookup (session_bus_entries, session);
-        if (entry)
-            g_dbus_method_invocation_return_value (invocation, g_variant_new ("(o)", entry->path));
-        else // FIXME: Need to make proper error
-            g_dbus_method_invocation_return_error_literal (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED, "Unable to find session for cookie");
-    }
 }
 
 static GVariant *
@@ -490,6 +418,34 @@ handle_seat_call (GDBusConnection       *connection,
         seat_switch_to_guest (seat, session_name);
         g_dbus_method_invocation_return_value (invocation, NULL);
     }
+    else if (g_strcmp0 (method_name, "Lock") == 0)
+    {
+        /* FIXME: Should only allow locks if have a session on this seat */
+        seat_lock (seat, NULL);
+        g_dbus_method_invocation_return_value (invocation, NULL);
+    }
+}
+
+static Seat *
+get_seat_for_session (Session *session)
+{
+    GList *seat_link;
+
+    for (seat_link = display_manager_get_seats (display_manager); seat_link; seat_link = seat_link->next)
+    {
+        Seat *seat = seat_link->data;
+        GList *display_link;
+
+        for (display_link = seat_get_displays (seat); display_link; display_link = display_link->next)
+        {
+            Display *display = display_link->data;
+
+            if (display_get_session (display) == session)
+                return seat;
+        }
+    } 
+
+    return NULL;
 }
 
 static GVariant *
@@ -508,7 +464,7 @@ handle_session_get_property (GDBusConnection       *connection,
     if (g_strcmp0 (property_name, "Seat") == 0)
         return g_variant_new_object_path (entry ? entry->parent_path : "");
     else if (g_strcmp0 (property_name, "UserName") == 0)
-        return g_variant_new_string (user_get_name (session_get_user (session)));
+        return g_variant_new_string (session_get_username (session));
 
     return NULL;
 }
@@ -523,6 +479,17 @@ handle_session_call (GDBusConnection       *connection,
                      GDBusMethodInvocation *invocation,
                      gpointer               user_data)
 {
+    Session *session = user_data;
+
+    if (g_strcmp0 (method_name, "Lock") == 0)
+    {
+        Seat *seat;
+
+        seat = get_seat_for_session (session);
+        /* FIXME: Should only allow locks if have a session on this seat */
+        seat_lock (seat, session_get_username (session));
+        g_dbus_method_invocation_return_value (invocation, NULL);
+    }
 }
 
 static BusEntry *
@@ -568,13 +535,13 @@ start_session_cb (Display *display, Seat *seat)
 
     session = display_get_session (display);
 
+    /* Set environment variables when session runs */
     seat_entry = g_hash_table_lookup (seat_bus_entries, seat);
     session_set_env (session, "XDG_SEAT_PATH", seat_entry->path);
-
     path = g_strdup_printf ("/org/freedesktop/DisplayManager/Session%d", session_index);
     session_index++;
     session_set_env (session, "XDG_SESSION_PATH", path);
-    g_free (path);
+    g_object_set_data_full (G_OBJECT (session), "XDG_SESSION_PATH", path, g_free);
 
     return FALSE;
 }
@@ -602,7 +569,7 @@ session_started_cb (Display *display, Seat *seat)
     g_signal_connect (session, "stopped", G_CALLBACK (session_stopped_cb), seat);
 
     seat_entry = g_hash_table_lookup (seat_bus_entries, seat);
-    entry = bus_entry_new (session_get_env (session, "XDG_SESSION_PATH"), seat_entry ? seat_entry->path : NULL, "SessionRemoved");
+    entry = bus_entry_new (g_object_get_data (G_OBJECT (session), "XDG_SESSION_PATH"), seat_entry ? seat_entry->path : NULL, "SessionRemoved");
     g_hash_table_insert (session_bus_entries, g_object_ref (session), entry);
 
     g_debug ("Registering session with bus path %s", entry->path);
@@ -706,14 +673,6 @@ bus_acquired_cb (GDBusConnection *connection,
         "      <arg name='display-number' direction='in' type='i'/>"
         "      <arg name='seat' direction='out' type='o'/>"
         "    </method>"
-        "    <method name='GetSeatForCookie'>"
-        "      <arg name='cookie' direction='in' type='s'/>"
-        "      <arg name='seat' direction='out' type='o'/>"
-        "    </method>"
-        "    <method name='GetSessionForCookie'>"
-        "      <arg name='cookie' direction='in' type='s'/>"
-        "      <arg name='session' direction='out' type='o'/>"
-        "    </method>"
         "    <signal name='SeatAdded'>"
         "      <arg name='seat' type='o'/>"
         "    </signal>"
@@ -747,6 +706,7 @@ bus_acquired_cb (GDBusConnection *connection,
         "    <method name='SwitchToGuest'>"
         "      <arg name='session-name' direction='in' type='s'/>"
         "    </method>"
+        "    <method name='Lock'/>"    
         "  </interface>"
         "</node>";
     const gchar *session_interface =
@@ -754,6 +714,7 @@ bus_acquired_cb (GDBusConnection *connection,
         "  <interface name='org.freedesktop.DisplayManager.Session'>"
         "    <property name='Seat' type='o' access='read'/>"
         "    <property name='UserName' type='s' access='read'/>"
+        "    <method name='Lock'/>"
         "  </interface>"
         "</node>";
     GDBusNodeInfo *display_manager_info;
@@ -854,13 +815,8 @@ main (int argc, char **argv)
     gboolean explicit_config = FALSE;
     gboolean test_mode = FALSE;
     gchar *pid_path = "/var/run/lightdm.pid";
-    gchar *xserver_command = NULL;
-    gchar *passwd_path = NULL;
     gchar *xsessions_dir = NULL;
     gchar *xgreeters_dir = NULL;
-    gchar *greeter_session = NULL;
-    gchar *user_session = NULL;
-    gchar *session_wrapper = NULL;
     gchar *config_dir;
     gchar *log_dir = NULL;
     gchar *run_dir = NULL;
@@ -868,8 +824,6 @@ main (int argc, char **argv)
     gchar *default_log_dir = g_strdup (LOG_DIR);
     gchar *default_run_dir = g_strdup (RUN_DIR);
     gchar *default_cache_dir = g_strdup (CACHE_DIR);
-    gchar *minimum_vt = NULL;
-    gchar *minimum_display_number = NULL;
     gboolean show_version = FALSE;
     GOptionEntry options[] = 
     {
@@ -882,30 +836,9 @@ main (int argc, char **argv)
         { "test-mode", 0, 0, G_OPTION_ARG_NONE, &test_mode,
           /* Help string for command line --test-mode flag */
           N_("Run as unprivileged user, skipping things that require root access"), NULL },
-        { "passwd-file", 0, 0, G_OPTION_ARG_STRING, &passwd_path,
-          /* Help string for command line --use-passwd flag */
-          N_("Use the given password file for authentication (for testing, requires --test-mode)"), "FILE" },
         { "pid-file", 0, 0, G_OPTION_ARG_STRING, &pid_path,
           /* Help string for command line --pid-file flag */
           N_("File to write PID into"), "FILE" },
-        { "xserver-command", 0, 0, G_OPTION_ARG_STRING, &xserver_command,
-          /* Help string for command line --xserver-command flag */
-          N_("Command to run X servers"), "COMMAND" },
-        { "greeter-session", 0, 0, G_OPTION_ARG_STRING, &greeter_session,
-          /* Help string for command line --greeter-session flag */
-          N_("Greeter session"), "SESSION" },
-        { "user-session", 0, 0, G_OPTION_ARG_STRING, &user_session,
-          /* Help string for command line --user-session flag */
-          N_("User session"), "SESSION" },
-        { "session-wrapper", 0, 0, G_OPTION_ARG_STRING, &session_wrapper,
-          /* Help string for command line --session-wrapper flag */
-          N_("Session wrapper"), "SESSION" },
-        { "minimum-vt", 0, 0, G_OPTION_ARG_STRING, &minimum_vt,
-          /* Help string for command line --minimum-vt flag */
-          N_("Minimum VT to use for X servers"), "NUMBER" },
-        { "minimum-display-number", 0, 0, G_OPTION_ARG_STRING, &minimum_display_number,
-          /* Help string for command line --minimum-display-number flag */
-          N_("Minimum display number to use for X servers"), "NUMBER" },
         { "xsessions-dir", 0, 0, G_OPTION_ARG_STRING, &xsessions_dir,
           /* Help string for command line --xsessions-dir flag */
           N_("Directory to load X sessions from"), "DIRECTORY" },
@@ -920,15 +853,18 @@ main (int argc, char **argv)
           N_("Directory to store running state"), "DIRECTORY" },
         { "cache-dir", 0, 0, G_OPTION_ARG_STRING, &cache_dir,
           /* Help string for command line --cache-dir flag */
-          N_("Directory to cached information"), "DIRECTORY" },
+          N_("Directory to cache information"), "DIRECTORY" },
         { "version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
           /* Help string for command line --version flag */
           N_("Show release version"), NULL },
         { NULL }
     };
     GError *error = NULL;
- 
-    g_thread_init (NULL);
+
+    /* When lightdm starts sessions it needs to run itself in a new mode */
+    if (argc >= 2 && strcmp (argv[1], "--session-child") == 0)
+        return session_child_run (argc, argv);
+
     g_type_init ();
     loop = g_main_loop_new (NULL, FALSE);
 
@@ -991,13 +927,6 @@ main (int argc, char **argv)
         g_free (xserver_path);
     }
 
-    /* Don't allow to be run as root and use a password file (asking for danger!) */
-    if (getuid () == 0 && passwd_path)
-    {
-        g_printerr ("Only allowed to use --passwd-file when running with --test-mode.\n"); 
-        return EXIT_FAILURE;
-    }
-
     /* Make sure the system binary directory (where the greeters are installed) is in the path */
     if (test_mode)
     {
@@ -1068,6 +997,8 @@ main (int argc, char **argv)
         config_set_boolean (config_get_instance (), "SeatDefaults", "start-session", TRUE);
     if (!config_has_key (config_get_instance (), "SeatDefaults", "allow-guest"))
         config_set_boolean (config_get_instance (), "SeatDefaults", "allow-guest", TRUE);
+    if (!config_has_key (config_get_instance (), "SeatDefaults", "greeter-allow-guest"))
+        config_set_boolean (config_get_instance (), "SeatDefaults", "greeter-allow-guest", TRUE);
     if (!config_has_key (config_get_instance (), "SeatDefaults", "greeter-session"))
         config_set_string (config_get_instance (), "SeatDefaults", "greeter-session", GREETER_SESSION);
     if (!config_has_key (config_get_instance (), "SeatDefaults", "user-session"))
@@ -1089,12 +1020,6 @@ main (int argc, char **argv)
         config_set_string (config_get_instance (), "LightDM", "xgreeters-directory", XGREETERS_DIR);
 
     /* Override defaults */
-    if (minimum_vt)
-        config_set_integer (config_get_instance (), "LightDM", "minimum-vt", atoi (minimum_vt));
-    g_free (minimum_vt);
-    if (minimum_display_number)
-        config_set_integer (config_get_instance (), "LightDM", "minimum-display-number", atoi (minimum_display_number));
-    g_free (minimum_display_number);
     if (log_dir)
         config_set_string (config_get_instance (), "LightDM", "log-directory", log_dir);
     g_free (log_dir);
@@ -1110,18 +1035,6 @@ main (int argc, char **argv)
     if (xgreeters_dir)
         config_set_string (config_get_instance (), "LightDM", "xgreeters-directory", xgreeters_dir);
     g_free (xgreeters_dir);
-    if (xserver_command)
-        config_set_string (config_get_instance (), "SeatDefaults", "xserver-command", xserver_command);
-    g_free (xserver_command);
-    if (greeter_session)
-        config_set_string (config_get_instance (), "SeatDefaults", "greeter-session", greeter_session);
-    g_free (greeter_session);
-    if (user_session)
-        config_set_string (config_get_instance (), "SeatDefaults", "user-session", user_session);
-    g_free (user_session);
-    if (session_wrapper)
-        config_set_string (config_get_instance (), "SeatDefaults", "session-wrapper", session_wrapper);
-    g_free (session_wrapper);
 
     /* Create run and cache directories */
     dir = config_get_string (config_get_instance (), "LightDM", "log-directory");
@@ -1153,12 +1066,6 @@ main (int argc, char **argv)
 
     if (getuid () != 0)
         g_debug ("Running in user mode");
-    if (passwd_path)
-    {
-        g_debug ("Using password file '%s' for authentication", passwd_path);
-        user_set_use_passwd_file (passwd_path);
-        pam_session_set_use_passwd_file (passwd_path);
-    }
     if (getenv ("DISPLAY"))
         g_debug ("Using Xephyr for X servers");
 
